from .. import click
from .build_wheels import build_wheels
from .compile_all import compile_all_requirements


@click.command()
def main():
    """
    Build wheels and compile all requirements.
    """
    build_wheels()
    compile_all_requirements()



import io
import os
import re
from glob import glob

from .. import click
from ._utils import call

WHEEL_DIR = 'wheels'

WHEEL_LINE_RX = re.compile(
    '^\s*(?P<pkg>(\w|-)+)(?P<verspec>\S*)\s+'
    '#\s*wheel:\s*(?P<wheel_spec>\S.*)$')


@click.command()
def main():
    """
    Build wheels of required packages.
    """
    build_wheels()


def build_wheels():
    to_build = list(get_packages_to_build())
    for (pkg, ver, url) in to_build:
        build_wheel(pkg, ver, url)


def get_packages_to_build():
    for requirement_file in glob('requirements*.in'):
        with io.open(requirement_file, 'rt', encoding='utf-8') as fp:
            for line in fp:
                if line.lstrip().startswith('#'):
                    continue
                m = WHEEL_LINE_RX.match(line)
                if m:
                    yield parse_requirement_wheel_line_data(**m.groupdict())


def parse_requirement_wheel_line_data(pkg, verspec, wheel_spec):
    if not verspec.startswith(('>=', '~=', '==')):
        raise Exception('No version (>=, ~=, ==) for: {}'.format(pkg))
    ver = verspec[2:]
    (provider, src) = wheel_spec.strip().split(':', 1)
    if provider == 'github':
        url = 'git+ssh://git@github.com/{}@v{}'.format(src, ver)
    else:
        raise Exception('Currently only github repos are supported')
    return (pkg, ver, url)


def build_wheel(pkg, ver, url):
    d = locals()
    if get_wheels(pkg, ver):
        print('*** Already built: {}'.format(get_wheels(pkg, ver)[0]))
        return
    print('*** Building wheel for {} {} from {}'.format(pkg, ver, url))
    call('pip wheel -v -w {w} --no-deps {u}', w=WHEEL_DIR, u=url)
    built_wheel = get_wheels(pkg, ver)[0]
    print('*** Built: {}'.format(built_wheel))
    for wheel in get_wheels(pkg):  # All versions
        if wheel != built_wheel:
            print('*** Removing: {}'.format(wheel))
            os.remove(wheel)


def get_wheels(pkg, ver='*'):
    return glob(os.path.join(
        WHEEL_DIR, '{}-{}-*.whl'.format(pkg.replace('-', '_'), ver)))



import collections
import os
import sys
from subprocess import check_call

from . import click
from .exceptions import IncompatibleRequirements, UnsupportedConstraint
from .utils import flat_map, key_from_req, is_pinned_requirement, is_vcs_link

PACKAGES_TO_IGNORE = [
    'pip',
    'prequ',
    'pip-review',
    'setuptools',
    'wheel',
    'pkg-resources',
]


def dependency_tree(installed_keys, root_key):
    """
    Calculate the dependency tree for the package `root_key` and return
    a collection of all its dependencies.  Uses a DFS traversal algorithm.

    `installed_keys` should be a {key: requirement} mapping, e.g.
        {'django': from_line('django==1.8')}
    `root_key` should be the key to return the dependency tree for.
    """
    dependencies = set()
    queue = collections.deque()

    if root_key in installed_keys:
        dep = installed_keys[root_key]
        queue.append(dep)

    while queue:
        v = queue.popleft()
        key = key_from_req(v)
        if key in dependencies:
            continue

        dependencies.add(key)

        for dep_specifier in v.requires():
            dep_name = key_from_req(dep_specifier)
            if dep_name in installed_keys:
                dep = installed_keys[dep_name]

                if dep_specifier.specifier.contains(dep.version):
                    queue.append(dep)

    return dependencies


def get_dists_to_ignore(installed):
    """
    Returns a collection of package names to ignore when performing prequ sync,
    based on the currently installed environment.  For example, when Prequ
    is installed in the local environment, it should be ignored, including all
    of its dependencies (e.g. click).  When Prequ is not installed
    locally, click should also be installed/uninstalled depending on the given
    requirements.
    """
    installed_keys = {key_from_req(r): r for r in installed}
    return list(flat_map(lambda req: dependency_tree(installed_keys, req), PACKAGES_TO_IGNORE))


def merge(requirements, ignore_conflicts):
    by_key = {}

    for ireq in requirements:
        if ((is_vcs_link(ireq) and not ireq.editable and not is_pinned_requirement(ireq))):
            msg = 'Prequ does not support non-editable vcs URLs that are not pinned to one version.'
            raise UnsupportedConstraint(msg, ireq)

        key = ireq.link or key_from_req(ireq.req)

        if not ignore_conflicts:
            existing_ireq = by_key.get(key)
            if existing_ireq:
                # NOTE: We check equality here since we can assume that the
                # requirements are all pinned
                if ireq.specifier != existing_ireq.specifier:
                    raise IncompatibleRequirements(ireq, existing_ireq)

        # TODO: Always pick the largest specifier in case of a conflict
        by_key[key] = ireq

    return by_key.values()


def diff(compiled_requirements, installed_dists):
    """
    Calculate which packages should be installed or uninstalled, given a set
    of compiled requirements and a list of currently installed modules.
    """
    requirements_lut = {r.link if r.editable else key_from_req(r.req): r for r in compiled_requirements}

    satisfied = set()  # holds keys
    to_install = set()  # holds keys-and-versions
    to_uninstall = set()  # holds keys

    pkgs_to_ignore = get_dists_to_ignore(installed_dists)
    for dist in installed_dists:
        key = key_from_req(dist)
        if key not in requirements_lut:
            to_uninstall.add(key)
        elif requirements_lut[key].specifier.contains(dist.version):
            satisfied.add(key)

    for key, requirement in requirements_lut.items():
        if key not in satisfied:
            to_install.add(str(requirement.link or requirement.req))

    # Make sure to not uninstall any packages that should be ignored
    to_uninstall -= set(pkgs_to_ignore)

    return (to_install, to_uninstall)


def sync(to_install, to_uninstall, verbose=False, dry_run=False, pip_flags=None, install_flags=None):
    """
    Install and uninstalls the given sets of modules.
    """
    if not to_uninstall and not to_install:
        click.echo("Everything up-to-date")

    if pip_flags is None:
        pip_flags = []

    if not verbose:
        pip_flags += ['-q']

    if os.environ.get('VIRTUAL_ENV'):
        # find pip via PATH
        pip = 'pip'
    else:
        # find pip in same directory as prequ entry-point script
        pip = os.path.join(os.path.dirname(os.path.abspath(sys.argv[0])), 'pip')

    if to_uninstall:
        if dry_run:
            click.echo("Would uninstall:")
            for pkg in to_uninstall:
                click.echo("  {}".format(pkg))
        else:
            check_call([pip, 'uninstall', '-y'] + pip_flags + sorted(to_uninstall))

    if to_install:
        if install_flags is None:
            install_flags = []
        if dry_run:
            click.echo("Would install:")
            for pkg in to_install:
                click.echo("  {}".format(pkg))
        else:
            check_call([pip, 'install'] + pip_flags + install_flags + sorted(to_install))
    return 0



from setuptools import find_packages, setup


def get_long_description():
    with open('README.rst') as fp:
        return fp.read()


setup(
    name='prequ',
    version='0.180.8',
    url='https://github.com/suutari-ai/prequ/',
    license='BSD',
    maintainer='Tuomas Suutari',
    maintainer_email='tuomas.suutari@anders.fi',
    description="Prequ -- Python requirement handling",
    long_description=get_long_description(),
    packages=find_packages(exclude=['tests']),
    install_requires=[
        'click>=6',
        'first',
        'six',
    ],
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'prequ = prequ.scripts.prequ:main',
        ],
    },
    platforms='any',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Topic :: System :: Systems Administration',
    ]
)



import subprocess

DEVNULL = subprocess.DEVNULL


def call(cmd, stdout=None, **kwargs):
    formatted_cmd = [x.format(**kwargs) for x in cmd.split()]
    return subprocess.check_call(formatted_cmd, stdout=stdout)



from .. import click
from . import build_wheels
from . import compile
from . import compile_all
from . import sync
from . import update


@click.group()
@click.version_option()
def main():
    pass


main.add_command(build_wheels.main, 'build-wheels')
main.add_command(compile.cli, 'compile')
main.add_command(compile_all.main, 'compile-all')
main.add_command(sync.cli, 'sync')
main.add_command(update.main, 'update')


if __name__ == '__main__':
    main()



import io
from glob import glob

from .. import click
from ._utils import call, DEVNULL


@click.command()
def main():
    """
    Compile all requirements files (*.in).
    """
    compile_all_requirements()


def compile_all_requirements():
    option_lines = list(get_option_lines('requirements.in'))
    src_reqs = ['requirements.in'] + sorted(glob('requirements-*.in'))
    for src_file in src_reqs:
        txt_file = src_file.replace('.in', '.txt')
        print('*** Compiling {} to {}'.format(src_file, txt_file))
        call('prequ compile --no-index --no-header --no-annotate {r}',
             r=src_file, stdout=DEVNULL)
        postprocess_requirement_file(txt_file, option_lines)


def get_option_lines(filename):
    ignored = (
        '-r', '--requirement',
        '-e', '--editable',
        '-c', '--constraint',
    )
    with open(filename, 'rt', encoding='utf-8') as fp:
        for line in fp:
            cleaned = line.strip()
            if cleaned.startswith('-') and not cleaned.startswith(ignored):
                yield cleaned


def postprocess_requirement_file(filename, option_lines):
    header = [
        '# This file is autogenerated by Prequ.  To update, run:',
        '#',
        '#   prequ update',
        '',
    ]
    with io.open(filename, 'r+t', encoding='utf-8') as fp:
        req_lines = [line.strip() for line in fp.readlines()]
        fp.seek(0)
        fp.truncate()
        for line in header + option_lines + [''] + req_lines:
            fp.write(line + '\n')
